from typing import Optional

from haystack import Pipeline
from haystack.core.errors import PipelineError


class _PipelineRegistry:
    def __init__(self) -> None:
        self._pipelines: dict[str, Pipeline] = {}

    def add(self, name: str, source: str) -> Pipeline:
        if name in self._pipelines:
            msg = f"A pipeline with name {name} is already in the registry."
            raise ValueError(msg)

        try:
            self._pipelines[name] = Pipeline.loads(source)
        except PipelineError as e:
            msg = f"Unable to parse Haystack Pipeline {name}: {e}"
            raise ValueError(msg) from e

        return self._pipelines[name]

    def remove(self, name: str):
        if name in self._pipelines:
            del self._pipelines[name]

    def get(self, name: str) -> Optional[Pipeline]:
        return self._pipelines.get(name)

    def get_names(self) -> list[str]:
        return list(self._pipelines.keys())


registry = _PipelineRegistry()



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0
import click

from hayhooks.cli.run import run
from hayhooks.cli.deploy import deploy
from hayhooks.cli.status import status
from hayhooks.cli.undeploy import undeploy


@click.group(context_settings={"help_option_names": ["-h", "--help"]}, invoke_without_command=True)
@click.version_option(prog_name="Hayhooks")
@click.option('--server', default="http://localhost:1416")
@click.pass_context
def hayhooks(ctx, server):
    ctx.obj = server


hayhooks.add_command(run)
hayhooks.add_command(deploy)
hayhooks.add_command(status)
hayhooks.add_command(undeploy)



from hayhooks.server import app
from hayhooks.server.utils.deploy_utils import deploy_pipeline_def, PipelineDefinition


@app.post("/deploy")
async def deploy(pipeline_def: PipelineDefinition):
    return deploy_pipeline_def(app, pipeline_def)



from pathlib import Path

import click
import requests


@click.command()
@click.pass_obj
@click.option('-n', '--name')
@click.argument('pipeline_file', type=click.File('r'))
def deploy(server, name, pipeline_file):
    if name is None:
        name = Path(pipeline_file.name).stem
    resp = requests.post(f"{server}/deploy", json={"name": name, "source_code": str(pipeline_file.read())})

    if resp.status_code >= 400:
        click.echo(f"Error deploying pipeline: {resp.json().get('detail')}")
    else:
        click.echo(f"Pipeline successfully deployed with name: {resp.json().get('name')}")



from typing import get_args, get_origin, List

from pydantic import BaseModel, create_model, ConfigDict
from haystack.dataclasses import Document


class HaystackDocument(BaseModel):
    id: str
    content: str


class PipelineDefinition(BaseModel):
    name: str
    source_code: str


def get_request_model(pipeline_name: str, pipeline_inputs):
    """
    Inputs have this form:
    {
        'first_addition': { <-- Component Name
            'value': {'type': <class 'int'>, 'is_mandatory': True}, <-- Input
            'add': {'type': typing.Optional[int], 'is_mandatory': False, 'default_value': None}, <-- Input
        },
        'second_addition': {'add': {'type': typing.Optional[int], 'is_mandatory': False}},
    }
    """
    request_model = {}
    config = ConfigDict(arbitrary_types_allowed=True)

    for component_name, inputs in pipeline_inputs.items():

        component_model = {}
        for name, typedef in inputs.items():
            component_model[name] = (typedef["type"], typedef.get("default_value", ...))
        request_model[component_name] = (create_model('ComponentParams', **component_model, __config__=config), ...)

    return create_model(f'{pipeline_name.capitalize()}RunRequest', **request_model, __config__=config)


def get_response_model(pipeline_name: str, pipeline_outputs):
    """
    Outputs have this form:
    {
        'second_addition': { <-- Component Name
            'result': {'type': "<class 'int'>"}  <-- Output
        },
    }
    """
    response_model = {}
    config = ConfigDict(arbitrary_types_allowed=True)

    for component_name, outputs in pipeline_outputs.items():
        component_model = {}
        for name, typedef in outputs.items():
            output_type = typedef["type"]
            if get_origin(output_type) == list and get_args(output_type)[0] == Document:
                component_model[name] = (List[HaystackDocument], ...)
            else:
                component_model[name] = (typedef["type"], ...)
        response_model[component_name] = (create_model('ComponentParams', **component_model, __config__=config), ...)

    return create_model(f'{pipeline_name.capitalize()}RunResponse', **response_model, __config__=config)


def convert_component_output(component_output):
    """
    Component output has this form:

    "documents":[
        {"id":"818170...", "content":"RapidAPI for Mac is a full-featured HTTP client."}
    ]

    We inspect the output and convert haystack.Document into the HaystackDocument pydantic model as needed
    """
    result = {}
    for output_name, data in component_output.items():
        # Empty containers, None values, empty strings and the likes: do nothing
        if not data:
            result[output_name] = data

        # Output contains a list of Document
        if type(data) is list and type(data[0]) is Document:
            result[output_name] = [HaystackDocument(id=d.id, content=d.content) for d in data]
        # Output is a single Document
        elif type(data) is Document:
            result[output_name] = HaystackDocument(id=data.id, content=data.content or "")
        # Any other type: do nothing
        else:
            result[output_name] = data

    return result



import click
import uvicorn
import os
import sys


@click.command()
@click.option('--host', default="localhost")
@click.option('--port', default=1416)
@click.option('--pipelines-dir', default=os.environ.get("HAYHOOKS_PIPELINES_DIR"))
@click.option('--additional-python-path', default=os.environ.get("HAYHOOKS_ADDITIONAL_PYTHONPATH"))
def run(host, port, pipelines_dir, additional_python_path):
    if not pipelines_dir:
        pipelines_dir = "pipelines.d"
    os.environ["HAYHOOKS_PIPELINES_DIR"] = pipelines_dir

    if additional_python_path:
        sys.path.append(additional_python_path)

    uvicorn.run("hayhooks.server:app", host=host, port=port)



import tempfile
from pathlib import Path

from fastapi import HTTPException
from fastapi.responses import FileResponse
from hayhooks.server import app
from hayhooks.server.pipelines import registry


@app.get("/draw/{pipeline_name}")
async def status(pipeline_name):
    pipeline = registry.get(pipeline_name)
    if not pipeline:
        raise HTTPException(status_code=404)

    _, fpath = tempfile.mkstemp()
    pipeline.draw(Path(fpath))
    return FileResponse(fpath, media_type="image/png")



# SPDX-FileCopyrightText: 2023-present Massimiliano Pippi <mpippi@gmail.com>
#
# SPDX-License-Identifier: Apache-2.0
import sys

if __name__ == "__main__":
    from hayhooks.cli import hayhooks

    sys.exit(hayhooks())
